/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { EMPTY_STRING } from '@alfa-client/tech-shared';
import { registerLocaleData } from '@angular/common';
import localeDe from '@angular/common/locales/de';
import localeDeExtra from '@angular/common/locales/extra/de';
import { NavigationE2EComponent } from 'apps/alfa-e2e/src/components/navigation/navigation.e2e.component';
import { VorgangListE2EComponent } from 'apps/alfa-e2e/src/components/vorgang/vorgang-list.e2e.component';
import { VorgangViewsE2EComponent } from 'apps/alfa-e2e/src/components/vorgang/vorgang-views.e2e.component';
import { HttpMethodE2E } from 'apps/alfa-e2e/src/model/util';
import { ClientAttributeNameE2E, VorgangE2E } from 'apps/alfa-e2e/src/model/vorgang';
import { MainPage, waitForSpinnerToDisappear } from 'apps/alfa-e2e/src/page-objects/main.po';
import { dropCollections, intercept, scrollToWindowBottom, waitOfInterceptor } from 'apps/alfa-e2e/src/support/cypress-helper';
import { exist } from 'apps/alfa-e2e/src/support/cypress.util';
import { createDateToday, generateIds } from 'apps/alfa-e2e/src/support/tech.util';
import { getUserSabineId, initUsermanagerUsers, loginAsSabine } from 'apps/alfa-e2e/src/support/user-util';
import {
  VORGANG_LIST_NEXT_FRIST_PARAM,
  VORGANG_LIST_PAGE_SIZE,
  buildVorgang,
  buildVorgangListPageLink,
  createAlfaClientAttributes,
  createNextWiedervorlageFristClientAttribute,
  initVorgaenge,
} from 'apps/alfa-e2e/src/support/vorgang-util';
import { Interception } from 'cypress/types/net-stubbing';

registerLocaleData(localeDe, 'de', localeDeExtra);

describe('VorgangList View Wiedervorlagen', () => {
  const mainPage: MainPage = new MainPage();

  const vorgangList: VorgangListE2EComponent = mainPage.getVorgangList();

  const navigation: NavigationE2EComponent = mainPage.getNavigation();

  const views: VorgangViewsE2EComponent = mainPage.getViews();

  const assignedTo: string = getUserSabineId();

  const ALLE_NAME_PREFIX: string = 'VorgangWithOpenWiedervorlagen_';
  const MEINE_VORGAENGE_NAME_PREFIX: string = 'VorgangWithOpenWiedervorlagenAssigned_';

  const ALLE_NAME: string = `${ALLE_NAME_PREFIX}_`;
  const MEINE_VORGAENGE_NAME: string = `${MEINE_VORGAENGE_NAME_PREFIX}_`;

  before(() => {
    const vorgaenge: VorgangE2E[] = createVorgaenge(ALLE_NAME_PREFIX, (id: string, name: string) =>
      buildVorgangWithOpenWiedervorlagen(id, name),
    ).concat(
      createVorgaenge(MEINE_VORGAENGE_NAME_PREFIX, (id: string, name: string) =>
        buildAssignedVorgangWithOpenWiedervorlagen(id, name),
      ),
    );
    initVorgaenge(vorgaenge);
    initUsermanagerUsers();

    loginAsSabine();
    exist(vorgangList.getRoot());
    waitForSpinnerToDisappear();
  });

  after(() => {
    dropCollections();
  });

  function createVorgaenge(namePrefix: string, buildFunction: (id: string, name: string) => VorgangE2E): VorgangE2E[] {
    let count = 1;
    return generateIds(110).map((randomId) => {
      const vorgang: VorgangE2E = buildFunction(randomId, `${namePrefix}_${count}`);
      count++;
      return vorgang;
    });
  }

  function buildVorgangWithOpenWiedervorlagen(id: string, name: string): VorgangE2E {
    return {
      ...buildVorgang(id, name),
      clientAttributes: createAlfaClientAttributes(
        ClientAttributeNameE2E.NEXT_WIEDERVORLAGE_FRIST,
        createNextWiedervorlageFristClientAttribute(createDateToday()),
      ),
    };
  }

  function buildAssignedVorgangWithOpenWiedervorlagen(id: string, name: string): VorgangE2E {
    return {
      ...buildVorgang(id, name),
      assignedTo,
      clientAttributes: createAlfaClientAttributes(
        ClientAttributeNameE2E.NEXT_WIEDERVORLAGE_FRIST,
        createNextWiedervorlageFristClientAttribute(createDateToday()),
      ),
    };
  }

  describe('Alle filter', () => {
    describe('1 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        navigation.getAlleFilter().click();
        waitForSpinnerToDisappear();
        intercept(HttpMethodE2E.GET, buildVorgangWithOpenWiedervorlagenPageLink(1)).as(loadPageInterceptor);
      });

      it('should show 1 page with 100 vorgaenge with open wiedervorlagen', () => {
        views.getWiedervorlagen().getRoot().click();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === VORGANG_LIST_PAGE_SIZE);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, ALLE_NAME + 1);
          assert(interception.response.body._embedded.vorgangHeaderList[99].name, ALLE_NAME + VORGANG_LIST_PAGE_SIZE);
        });
      });
    });

    describe('2 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        intercept(HttpMethodE2E.GET, buildVorgangWithOpenWiedervorlagenPageLink(2)).as(loadPageInterceptor);
      });

      it('should show 2 page with 10 vorgaenge with open wiedervorlagen and 90 vorgaenge with open wiedervorlagen ASSIGNED on scrolling to bottom', () => {
        scrollToWindowBottom();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === VORGANG_LIST_PAGE_SIZE);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, MEINE_VORGAENGE_NAME + 1);
          assert(interception.response.body._embedded.vorgangHeaderList[99].name, MEINE_VORGAENGE_NAME + VORGANG_LIST_PAGE_SIZE);
        });
      });
    });

    describe('3 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        intercept(HttpMethodE2E.GET, buildVorgangWithOpenWiedervorlagenPageLink(3)).as(loadPageInterceptor);
      });

      it('should show 3 page with 10 vorgaenge with open wiedervorlagen ASSIGNED on scrolling to bottom', () => {
        scrollToWindowBottom();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === 20);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, MEINE_VORGAENGE_NAME + 80);
          assert(interception.response.body._embedded.vorgangHeaderList[19].name, MEINE_VORGAENGE_NAME + VORGANG_LIST_PAGE_SIZE);
        });
      });
    });

    function buildVorgangWithOpenWiedervorlagenPageLink(page: number): string {
      return `${buildVorgangListPageLink(page)}&${VORGANG_LIST_NEXT_FRIST_PARAM}`;
    }
  });

  describe('Meine Vorgänge filter', () => {
    describe('1 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        intercept(HttpMethodE2E.GET, buildVorgangWithOpenWiedervorlagenAssignedPageLink(1)).as(loadPageInterceptor);
      });

      it('should show 1 page with 100 vorgaenge with open wiedervorlagen', () => {
        navigation.getMeineVorgaengeFilter().click();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === VORGANG_LIST_PAGE_SIZE);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, MEINE_VORGAENGE_NAME + 1);
          assert(interception.response.body._embedded.vorgangHeaderList[99].name, MEINE_VORGAENGE_NAME + VORGANG_LIST_PAGE_SIZE);
        });
      });
    });

    describe('2 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        intercept(HttpMethodE2E.GET, buildVorgangWithOpenWiedervorlagenAssignedPageLink(2)).as(loadPageInterceptor);
      });

      it('should show 2 page with 10 vorgaenge with open wiedervorlagen and 90 vorgaenge with open wiedervorlagen ASSIGNED on scrolling to bottom', () => {
        scrollToWindowBottom();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === 10);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, MEINE_VORGAENGE_NAME + 101);
          assert(interception.response.body._embedded.vorgangHeaderList[9].name, MEINE_VORGAENGE_NAME + 110);
        });
      });
    });

    function buildVorgangWithOpenWiedervorlagenAssignedPageLink(page: number): string {
      return `${buildVorgangListPageLink(page, assignedTo)}&${VORGANG_LIST_NEXT_FRIST_PARAM}`;
    }
  });

  describe('Nicht zugewiesen', () => {
    describe('1 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        intercept(HttpMethodE2E.GET, buildUnassignedVorgangWithOpenWiedervorlagenPageLink(1)).as(loadPageInterceptor);
      });

      it('should show 1 page with 100 vorgaenge with open wiedervorlagen', () => {
        navigation.getUnassignedFilter().click();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === VORGANG_LIST_PAGE_SIZE);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, ALLE_NAME + 1);
          assert(interception.response.body._embedded.vorgangHeaderList[99].name, ALLE_NAME + VORGANG_LIST_PAGE_SIZE);
        });
      });
    });

    describe('2 page', () => {
      const loadPageInterceptor: string = 'loadPage';

      beforeEach(() => {
        intercept(HttpMethodE2E.GET, buildUnassignedVorgangWithOpenWiedervorlagenPageLink(2)).as(loadPageInterceptor);
      });

      it('should show 2 page with 10 vorgaenge on scrolling to bottom', () => {
        scrollToWindowBottom();

        waitOfInterceptor(loadPageInterceptor).then((interception: Interception) => {
          assert(interception.response.body._embedded.vorgangHeaderList.length === 10);
          assert(interception.response.body._embedded.vorgangHeaderList[0].name, ALLE_NAME + 101);
          assert(interception.response.body._embedded.vorgangHeaderList[9].name, ALLE_NAME + 110);
        });
      });
    });

    function buildUnassignedVorgangWithOpenWiedervorlagenPageLink(page: number): string {
      return `${buildVorgangListPageLink(page, EMPTY_STRING)}&${VORGANG_LIST_NEXT_FRIST_PARAM}`;
    }
  });
});
