/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { NavigationService } from '@alfa-client/navigation-shared';
import { StateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { TestBed } from '@angular/core/testing';
import { singleColdCompleted } from 'libs/tech-shared/test/marbles';
import { Observable, of } from 'rxjs';
import {
  createAdminOrganisationsEinheitListResource,
  createAdminOrganisationsEinheitResource,
} from '../test/organisations-einheit';
import { OrganisationsEinheitListResourceService } from './organisations-einheit-list-resource.service';
import { OrganisationsEinheitResourceService } from './organisations-einheit-resource.service';
import { AdminOrganisationsEinheitListResource, AdminOrganisationsEinheitResource } from './organisations-einheit.model';
import { OrganisationsEinheitService } from './organisations-einheit.service';

jest.mock('./organisations-einheit-list-resource.service');
jest.mock('./organisations-einheit-resource.service');
jest.mock('libs/navigation-shared/src/lib/navigation.service');
jest.mock('libs/ui/src/lib/snackbar/snackbar.service');

describe('OrganisationsEinheitService', () => {
  let service: OrganisationsEinheitService;
  let listResourceService: Mock<OrganisationsEinheitListResourceService>;
  let resourceService: Mock<OrganisationsEinheitResourceService>;
  let navigationService: Mock<NavigationService>;
  let snackBarService: Mock<SnackBarService>;

  beforeEach(() => {
    listResourceService = mock(OrganisationsEinheitListResourceService);
    resourceService = mock(OrganisationsEinheitResourceService);

    navigationService = mock(NavigationService);
    navigationService.urlChanged.mockReturnValue(of({}));

    snackBarService = mock(SnackBarService);

    TestBed.configureTestingModule({
      providers: [
        { provide: OrganisationsEinheitListResourceService, useValue: useFromMock(listResourceService) },
        { provide: OrganisationsEinheitResourceService, useValue: useFromMock(resourceService) },
        { provide: NavigationService, useValue: useFromMock(navigationService) },
        { provide: SnackBarService, useValue: useFromMock(snackBarService) },
      ],
    });

    service = TestBed.inject(OrganisationsEinheitService);
  });

  describe('onNavigation', () => {
    it('should not call listResourceService select', () => {
      service.onNavigation({});

      expect(listResourceService.select).not.toHaveBeenCalled();
    });

    it('should call listResourceService select', () => {
      service.onNavigation({ [OrganisationsEinheitService.ORGANISATIONS_EINHEIT_URL]: 'some-uri' });

      expect(listResourceService.select).toHaveBeenCalled();
    });

    it('should call getOrganisationsEinheitUrl', () => {
      service.getOrganisationsEinheitUrl = jest.fn();

      service.onNavigation({ [OrganisationsEinheitService.ORGANISATIONS_EINHEIT_URL]: 'some-uri' });

      expect(service.getOrganisationsEinheitUrl).toHaveBeenCalled();
    });
  });

  describe('getOrganisationsEinheitUrl', () => {
    it('should call navigationService getDecodedParam', () => {
      service.getOrganisationsEinheitUrl();

      expect(navigationService.getDecodedParam).toHaveBeenCalled();
    });

    it('should return uri', () => {
      navigationService.getDecodedParam.mockReturnValue('some-uri');

      expect(service.getOrganisationsEinheitUrl()).toBe('some-uri');
    });
  });

  describe('getList', () => {
    const organisationsEinheitListResource: AdminOrganisationsEinheitListResource = createAdminOrganisationsEinheitListResource();
    const organisationsEinheitStateListResource: StateResource<AdminOrganisationsEinheitListResource> =
      createStateResource(organisationsEinheitListResource);

    beforeEach(() => {
      listResourceService.getList.mockReturnValue(of(organisationsEinheitStateListResource));
    });

    it('should call listResourceService', () => {
      service.getList();

      expect(listResourceService.getList).toHaveBeenCalled();
    });

    it('should return value', () => {
      const list$: Observable<StateResource<AdminOrganisationsEinheitListResource>> = service.getList();

      expect(list$).toBeObservable(singleColdCompleted(organisationsEinheitStateListResource));
    });
  });

  describe('get', () => {
    const organisationsEinheitResource: AdminOrganisationsEinheitResource = createAdminOrganisationsEinheitResource();
    const organisationsEinheitStateResource: StateResource<AdminOrganisationsEinheitResource> =
      createStateResource(organisationsEinheitResource);

    beforeEach(() => {
      resourceService.get.mockReturnValue(of(organisationsEinheitStateResource));
    });

    it('should call resourceService', () => {
      service.get();

      expect(resourceService.get).toHaveBeenCalled();
    });

    it('should return value', () => {
      const resource$: Observable<StateResource<AdminOrganisationsEinheitResource>> = service.get();

      expect(resource$).toBeObservable(singleColdCompleted(organisationsEinheitStateResource));
    });
  });

  describe('patch', () => {
    const organisationsEinheitResource: AdminOrganisationsEinheitResource = createAdminOrganisationsEinheitResource();
    const organisationsEinheitStateResource: StateResource<AdminOrganisationsEinheitResource> =
      createStateResource(organisationsEinheitResource);

    beforeEach(() => {
      resourceService.patch.mockReturnValue(of(organisationsEinheitStateResource));
    });

    it('should call resourceService', () => {
      service.patch(organisationsEinheitResource.settings).subscribe();

      expect(resourceService.patch).toHaveBeenCalled();
    });

    it('should call snackBarService showInfo', () => {
      service.patch(organisationsEinheitResource.settings).subscribe();

      expect(snackBarService.showInfo).toHaveBeenCalled();
    });

    it('should return value', () => {
      const resource$: Observable<StateResource<AdminOrganisationsEinheitResource>> = service.patch(
        organisationsEinheitResource.settings,
      );

      resource$.subscribe((result: StateResource<AdminOrganisationsEinheitResource>) => {
        expect(result).toEqual(organisationsEinheitStateResource);
      });
    });
  });
});
