/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { Mock, mock } from '@alfa-client/test-utils';
import { TestBed, fakeAsync, tick } from '@angular/core/testing';
import { faker } from '@faker-js/faker';
import KcAdminClient from '@keycloak/keycloak-admin-client';
import { TokenProvider } from '@keycloak/keycloak-admin-client/lib/client';
import GroupRepresentation from '@keycloak/keycloak-admin-client/lib/defs/groupRepresentation';
import MappingsRepresentation from '@keycloak/keycloak-admin-client/lib/defs/mappingsRepresentation';
import { RoleMappingPayload } from '@keycloak/keycloak-admin-client/lib/defs/roleRepresentation';
import { OAuthService } from 'angular-oauth2-oidc';
import { cold } from 'jest-marbles';
import { throwError } from 'rxjs';
import { UserAddFormservice } from '../../../user/src/lib/users-roles/user-add-form/user-add.formservice';
import { createUser } from '../../test/user';
import { User } from './user.model';
import { UserRepository } from './user.repository.service';

describe('UserRepository', () => {
  const accessToken: string = faker.string.alphanumeric(40);

  let repository: UserRepository;

  let kcAdminClient: Mock<KcAdminClient>;
  let oAuthService: Mock<OAuthService>;

  beforeEach(() => {
    kcAdminClient = mock(KcAdminClient);
    oAuthService = mock(OAuthService);
    TestBed.configureTestingModule({
      providers: [
        { provide: OAuthService, useValue: oAuthService },
        { provide: KcAdminClient, useValue: kcAdminClient },
      ],
    });
    oAuthService.getAccessToken.mockReturnValue(accessToken);
    repository = TestBed.inject(UserRepository);
  });

  it('should be created', () => {
    expect(repository).toBeTruthy();
  });

  describe('registerTokenProvider', () => {
    it('should register token provider from oauth service', async () => {
      const tokenProvider: TokenProvider = kcAdminClient.registerTokenProvider.mock.calls[0][0];
      const token: string = await tokenProvider.getAccessToken();
      expect(token).toEqual(accessToken);
    });
  });

  describe('createInKeycloak', () => {
    const user: User = createUser();

    beforeEach(() => {
      kcAdminClient.users = <any>{
        create: jest.fn().mockReturnValue(Promise.resolve({ id: user.id })),
        findOne: jest.fn().mockReturnValue(Promise.resolve(user)),
        addClientRoleMappings: jest.fn().mockReturnValue(Promise.resolve()),
        executeActionsEmail: jest.fn().mockReturnValue(Promise.resolve()),
      };
      kcAdminClient.clients = <any>{
        find: jest.fn().mockReturnValue(Promise.resolve([{ id: faker.string.uuid() }])),
        listRoles: jest.fn().mockReturnValue(Promise.resolve([{ id: faker.string.uuid(), name: faker.word.sample() }])),
      };
    });

    it('should call kcAdminClient users create', () => {
      repository.createInKeycloak(user);

      expect(kcAdminClient.users['create']).toBeCalledWith(user);
    });

    it('should call addUserRoles', fakeAsync(() => {
      const addUserRolesSpy: jest.SpyInstance = jest.spyOn(repository, 'addUserRoles');

      repository.createInKeycloak(user).subscribe();
      tick();

      expect(addUserRolesSpy).toBeCalledWith(user.id, user.clientRoles);
    }));

    it('should call sendActivationMail', (done) => {
      const sendActivationMailSpy: jest.SpyInstance = jest.spyOn(repository, 'sendActivationMail');
      jest.spyOn(repository, 'addUserRoles').mockReturnValue(Promise.resolve());

      repository.createInKeycloak(user).subscribe(() => {
        expect(sendActivationMailSpy).toBeCalledWith(user.id);
        done();
      });
    });

    it('should call getUserById', (done) => {
      const getUserByIdSpy: jest.SpyInstance = jest.spyOn(repository, 'getUserById');

      repository.createInKeycloak(user).subscribe(() => {
        expect(getUserByIdSpy).toBeCalledWith(user.id);
        done();
      });
    });

    it('should call handleError', fakeAsync(() => {
      const handleErrorSpy: jest.SpyInstance = jest.spyOn(repository, 'handleError');
      kcAdminClient.users['create'] = jest.fn().mockReturnValue(throwError(() => new Error('error')));

      repository.createInKeycloak(user).subscribe({ error: () => {} });
      tick();

      expect(handleErrorSpy).toHaveBeenCalled();
    }));

    describe('addUserRoles', () => {
      it('should call addUserRolesForClient for admin', async () => {
        const addUserRolesForClientSpy: jest.SpyInstance = jest.spyOn(repository, 'addUserRolesForClient');

        await repository.addUserRoles(user.id, { admin: [UserAddFormservice.ADMIN], alfa: [] });

        expect(addUserRolesForClientSpy).toBeCalledWith(user.id, [UserAddFormservice.ADMIN], UserRepository.ADMIN_CLIENT_NAME);
      });

      it('should call addUserRolesForClient for admin', async () => {
        const addUserRolesForClientSpy: jest.SpyInstance = jest.spyOn(repository, 'addUserRolesForClient');

        await repository.addUserRoles(user.id, { alfa: [UserAddFormservice.POSTSTELLE], admin: [] });

        expect(addUserRolesForClientSpy).toBeCalledWith(
          user.id,
          [UserAddFormservice.POSTSTELLE],
          UserRepository.ALFA_CLIENT_NAME,
        );
      });

      it('should not call addUserRolesForClient if clientRoles alfa and admin are empty', async () => {
        const addUserRolesForClientSpy: jest.SpyInstance = jest.spyOn(repository, 'addUserRolesForClient');

        await repository.addUserRoles(user.id, { admin: [], alfa: [] });

        expect(addUserRolesForClientSpy).not.toHaveBeenCalled();
      });
    });

    describe('addUserRolesForClient', () => {
      const clientId: string = faker.string.uuid();
      const roleMapping: RoleMappingPayload[] = [{ id: faker.string.uuid(), name: faker.word.sample() }];
      let getClientIdSpy: jest.SpyInstance;
      let mapUserRolesSpy: jest.SpyInstance;
      let addUserRolesInKeycloak: jest.SpyInstance;

      beforeEach(() => {
        getClientIdSpy = jest.spyOn(repository, 'getClientId').mockReturnValue(Promise.resolve(clientId));
        mapUserRolesSpy = jest.spyOn(repository, 'mapUserRoles').mockReturnValue(Promise.resolve(roleMapping));
        addUserRolesInKeycloak = jest.spyOn(repository, 'addUserRolesInKeycloak');
      });

      it('should call getClientId', async () => {
        await repository.addUserRolesForClient(user.id, [UserAddFormservice.ADMIN], UserRepository.ADMIN_CLIENT_NAME);

        expect(getClientIdSpy).toBeCalled();
      });

      it('should call getAlfaClientId', async () => {
        await repository.addUserRolesForClient(user.id, [UserAddFormservice.ADMIN], UserRepository.ADMIN_CLIENT_NAME);

        expect(mapUserRolesSpy).toHaveBeenCalledWith(clientId, [UserAddFormservice.ADMIN]);
      });

      it('should call addUserRolesInKeycloak', async () => {
        await repository.addUserRolesForClient(user.id, [UserAddFormservice.ADMIN], UserRepository.ADMIN_CLIENT_NAME);

        expect(addUserRolesInKeycloak).toHaveBeenCalledWith(user.id, clientId, roleMapping);
      });
    });

    describe('sendActivationMail', () => {
      it('should call kcAdminClient users executeActionsEmail', () => {
        const userId: string = faker.string.uuid();

        repository.sendActivationMail(userId);

        expect(kcAdminClient.users['executeActionsEmail']).toBeCalledWith({
          id: userId,
          actions: ['VERIFY_EMAIL'],
          lifespan: 3600 * 24 * 7,
        });
      });
    });

    describe('getUserById', () => {
      it('should call kcAdminClient users findOne', () => {
        const userId: string = faker.string.uuid();

        repository.getUserById(userId);

        expect(kcAdminClient.users['findOne']).toBeCalledWith({ id: userId });
      });
    });

    describe('handleError', () => {
      it('should throw error', () => {
        const error: Error = new Error('error');
        const result = repository.handleError(error);

        expect(result).toBeObservable(cold('#', null, new Error('An error occurred while creating the user.')));
      });
    });
  });

  describe('getUsers', () => {
    const userRep: User = createUser();
    const userRepArray: User[] = [userRep, userRep, userRep];
    const group: string = faker.word.sample();
    const alfaRole: string = faker.word.sample();
    const adminRole: string = faker.word.sample();
    const groupRep: GroupRepresentation[] = [{ name: group }];
    const roleRep: MappingsRepresentation = {
      clientMappings: {
        alfa: { mappings: [{ name: alfaRole }] },
        admin: { mappings: [{ name: adminRole }] },
      },
    };

    beforeEach(() => {
      jest.clearAllMocks();
      kcAdminClient.users = <any>{
        find: jest.fn().mockReturnValue(Promise.resolve(userRepArray)),
        listGroups: jest.fn().mockReturnValue(Promise.resolve(groupRep)),
        listRoleMappings: jest.fn().mockReturnValue(Promise.resolve(roleRep)),
      };
    });

    it('should call kcAdminClient users find', () => {
      repository.getUsers();

      expect(kcAdminClient.users['find']).toHaveBeenCalled();
    });

    it('should call mapToUser', fakeAsync(() => {
      const mapToUser: jest.SpyInstance<User> = jest.spyOn(repository, 'mapToUser');

      repository.getUsers().subscribe();
      tick();

      expect(mapToUser).toBeCalledTimes(userRepArray.length);
    }));

    it('should call kcadminClient listGroups for every user', fakeAsync(() => {
      repository.getUsers().subscribe();
      tick();

      expect(kcAdminClient.users['listGroups']).toBeCalledTimes(userRepArray.length);
    }));

    it('should call kcadminClient listRoleMappings for every user', fakeAsync(() => {
      repository.getUsers().subscribe();
      tick();

      expect(kcAdminClient.users['listRoleMappings']).toBeCalledTimes(userRepArray.length);
    }));

    it('should return users with groups and roles', (done) => {
      repository.getUsers().subscribe((users: User[]) => {
        users.forEach((user) =>
          expect(user).toEqual({ ...userRep, groups: [group], clientRoles: { alfa: [alfaRole], admin: [adminRole] } }),
        );
        done();
      });
    });

    it('should return users with empty groups and roles if they have none', (done) => {
      kcAdminClient.users['listGroups'] = jest.fn().mockReturnValue(Promise.resolve([]));
      kcAdminClient.users['listRoleMappings'] = jest.fn().mockReturnValue(Promise.resolve({}));

      repository.getUsers().subscribe((users: User[]) => {
        users.forEach((user) => expect(user).toEqual({ ...userRep, groups: [], clientRoles: { alfa: [], admin: [] } }));
        done();
      });
    });
  });
});
