/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { Injectable } from '@angular/core';
import KcAdminClient from '@keycloak/keycloak-admin-client';
import { TokenProvider } from '@keycloak/keycloak-admin-client/lib/client';
import ClientRepresentation from '@keycloak/keycloak-admin-client/lib/defs/clientRepresentation';
import GroupRepresentation from '@keycloak/keycloak-admin-client/lib/defs/groupRepresentation';
import MappingsRepresentation from '@keycloak/keycloak-admin-client/lib/defs/mappingsRepresentation';
import RoleRepresentation, { RoleMappingPayload } from '@keycloak/keycloak-admin-client/lib/defs/roleRepresentation';
import UserRepresentation from '@keycloak/keycloak-admin-client/lib/defs/userRepresentation';
import { OAuthService } from 'angular-oauth2-oidc';
import { isNil } from 'lodash-es';
import { Observable, catchError, concatMap, forkJoin, from, map, mergeMap, switchMap, tap, throwError } from 'rxjs';
import { ClientMapping, ClientRoles, User } from './user.model';

@Injectable({
  providedIn: 'root',
})
export class UserRepository {
  constructor(
    private oAuthService: OAuthService,
    private kcAdminClient: KcAdminClient,
  ) {
    this.registerAccessTokenProvider();
  }

  public static readonly ALFA_CLIENT_NAME: string = 'alfa';
  public static readonly ADMIN_CLIENT_NAME: string = 'admin';

  private registerAccessTokenProvider(): void {
    this.kcAdminClient.registerTokenProvider(this.getTokenProvider());
  }

  private getTokenProvider(): TokenProvider {
    return {
      getAccessToken: () => Promise.resolve(this.oAuthService.getAccessToken()),
    };
  }

  public createInKeycloak(user: Partial<User>): Observable<User> {
    return from(this.kcAdminClient.users.create(user)).pipe(
      concatMap(async (response: { id: string }) => {
        await this.addUserRoles(response.id, user.clientRoles);
        return response;
      }),
      tap((response: { id: string }) => this.sendActivationMail(response.id)),
      switchMap((response: { id: string }): Observable<User> => this.getUserById(response.id)),
      catchError((err) => this.handleError(err)),
    );
  }

  async addUserRoles(userId: string, clientRoles: ClientRoles): Promise<void> {
    if (clientRoles.admin.length > 0) {
      await this.addUserRolesForClient(userId, clientRoles.admin, UserRepository.ADMIN_CLIENT_NAME);
    }

    if (clientRoles.alfa.length > 0) {
      await this.addUserRolesForClient(userId, clientRoles.alfa, UserRepository.ALFA_CLIENT_NAME);
    }
  }

  async addUserRolesForClient(userId: string, userRoles: string[], client: string): Promise<void> {
    const clientId: string = await this.getClientId(client);
    const roles: RoleMappingPayload[] = await this.mapUserRoles(clientId, userRoles);
    await this.addUserRolesInKeycloak(userId, clientId, roles);
  }

  async getClientId(client: string): Promise<string | undefined> {
    const clients: ClientRepresentation[] = await this.kcAdminClient.clients.find({ clientId: client });
    return clients?.[0].id;
  }

  async mapUserRoles(clientId: string, userRoles: string[]): Promise<RoleMappingPayload[]> {
    const roles: RoleRepresentation[] = await this.kcAdminClient.clients.listRoles({ id: clientId });
    return roles.filter((role) => userRoles.includes(role.name)).map((role) => ({ id: role.id, name: role.name }));
  }

  async addUserRolesInKeycloak(userId: string, clientId: string, roles: RoleMappingPayload[]): Promise<void> {
    await this.kcAdminClient.users.addClientRoleMappings({
      id: userId,
      clientUniqueId: clientId,
      roles,
    });
  }

  sendActivationMail(userId: string): void {
    this.kcAdminClient.users.executeActionsEmail({
      id: userId,
      actions: ['VERIFY_EMAIL'],
      lifespan: 3600 * 24 * 7,
    });
  }

  getUserById(userId: string): Observable<User> {
    return from(this.kcAdminClient.users.findOne({ id: userId })) as Observable<User>;
  }

  handleError(err: any): Observable<never> {
    console.error('Error creating user in Keycloak:', err);
    return throwError(() => new Error('An error occurred while creating the user.'));
  }

  public getUsers(): Observable<User[]> {
    return from(this.kcAdminClient.users.find()).pipe(
      map((userReps: UserRepresentation[]) => userReps.map((userReps) => this.mapToUser(userReps))),
      mergeMap((users) => forkJoin(users.map((users) => this.addInformationToUser(users)))),
    );
  }

  mapToUser(userRepresentation: UserRepresentation): User {
    return {
      id: userRepresentation.id,
      email: userRepresentation.email,
      username: userRepresentation.username,
      firstName: userRepresentation.firstName,
      lastName: userRepresentation.lastName,
      enabled: userRepresentation.enabled,
      groups: null,
      clientRoles: {
        alfa: [],
        admin: [],
      },
    };
  }

  private addInformationToUser(user: User): Observable<User> {
    return forkJoin([this.getUserGroups(user), this.getClientRoles(user)]).pipe(
      map(([groups, clientRoles]) => ({
        ...user,
        groups,
        clientRoles,
      })),
    );
  }

  private getUserGroups(user: User): Observable<string[]> {
    return from(this.kcAdminClient.users.listGroups({ id: user.id })).pipe(
      map((groups: GroupRepresentation[]) => groups.map((group) => group.name)),
    );
  }

  private getClientRoles(user: User): Observable<ClientRoles> {
    return from(this.kcAdminClient.users.listRoleMappings({ id: user.id })).pipe(
      map((roleMappings: MappingsRepresentation) => {
        return {
          alfa: this.mapToClientRoleNames(roleMappings, UserRepository.ALFA_CLIENT_NAME),
          admin: this.mapToClientRoleNames(roleMappings, UserRepository.ADMIN_CLIENT_NAME),
        };
      }),
    );
  }

  private mapToClientRoleNames(roleMappings: MappingsRepresentation, client: string): string[] {
    const clientMappingsAlfa: ClientMapping | undefined = roleMappings?.clientMappings?.[client];

    if (isNil(clientMappingsAlfa)) {
      return [];
    }

    return clientMappingsAlfa.mappings.map((role: RoleRepresentation) => role.name);
  }
}
