/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { User, UserService } from '@admin-client/user-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import { createStateResource, StateResource } from '@alfa-client/tech-shared';
import { dispatchEventFromFixture, getDebugElementFromFixtureByCss, mock, Mock } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { CommonModule } from '@angular/common';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { ReactiveFormsModule } from '@angular/forms';
import { ButtonWithSpinnerComponent, CheckboxEditorComponent, TextEditorComponent } from '@ods/component';
import { cold } from 'jest-marbles';
import { createUser } from 'libs/admin/user-shared/test/user';
import { MockComponent } from 'ng-mocks';
import { of } from 'rxjs';
import { getDataTestIdAttributeOf } from '../../../../../../tech-shared/test/data-test';
import { UserAddFormComponent } from './user-add-form.component';
import { UserAddFormservice } from './user-add.formservice';

describe('UserAddFormComponent', () => {
  let component: UserAddFormComponent;
  let fixture: ComponentFixture<UserAddFormComponent>;

  let formService: UserAddFormservice;

  let userService: Mock<UserService>;
  let navigationService: Mock<NavigationService>;
  let snackBarService: Mock<SnackBarService>;

  const saveButton: string = getDataTestIdAttributeOf('save-button');

  beforeEach(async () => {
    userService = mock(UserService);
    navigationService = mock(NavigationService);
    snackBarService = mock(SnackBarService);

    await TestBed.configureTestingModule({
      imports: [CommonModule, ReactiveFormsModule],
      declarations: [
        UserAddFormComponent,
        MockComponent(ButtonWithSpinnerComponent),
        MockComponent(TextEditorComponent),
        MockComponent(CheckboxEditorComponent),
      ],
      providers: [
        UserAddFormservice,
        { provide: UserService, useValue: userService },
        { provide: NavigationService, useValue: navigationService },
        { provide: SnackBarService, useValue: snackBarService },
      ],
    }).compileComponents();

    fixture = TestBed.createComponent(UserAddFormComponent);
    component = fixture.componentInstance;
    formService = component.formService;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('component', () => {
    describe('submit', () => {
      const userStateResource: StateResource<User> = createStateResource(createUser());

      beforeEach(() => {
        formService.submit = jest.fn().mockReturnValue(of(userStateResource));
      });

      it('should call formService submit if form is valid', () => {
        jest.spyOn(formService.form, 'invalid', 'get').mockReturnValue(false);

        component.submit();

        expect(formService.submit).toHaveBeenCalled();
      });

      it('should not call formService submit if form is not valid', () => {
        jest.spyOn(formService.form, 'invalid', 'get').mockReturnValue(true);

        component.submit();

        expect(formService.submit).not.toHaveBeenCalled();
      });

      it('should set submitState$ if form is valid', () => {
        jest.spyOn(formService.form, 'invalid', 'get').mockReturnValue(false);

        component.submit();

        expect(component.submitStateResource$).toBeObservable(cold('(a|)', { a: userStateResource }));
      });
    });
  });

  describe('template', () => {
    describe('ods-button-with-spinner save', () => {
      describe('input', () => {
        it('should be set', () => {
          const stateResource: StateResource<User> = createStateResource(createUser());
          component.submitStateResource$ = of(stateResource);

          fixture.detectChanges();

          expect(getDebugElementFromFixtureByCss(fixture, saveButton).componentInstance.stateResource).toEqual(stateResource);
        });
      });

      describe('output', () => {
        it('should call submit', () => {
          const submitSpy: jest.SpyInstance = jest.spyOn(component, 'submit');

          dispatchEventFromFixture(fixture, saveButton, 'clickEmitter');

          expect(submitSpy).toHaveBeenCalled();
        });
      });
    });
  });
});
