/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { ROUTES } from '@admin-client/shared';
import { User, UserService } from '@admin-client/user-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import { createEmptyStateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, mock } from '@alfa-client/test-utils';
import { SnackBarService } from '@alfa-client/ui';
import { fakeAsync, TestBed, tick } from '@angular/core/testing';
import { AbstractControl, UntypedFormBuilder, UntypedFormGroup } from '@angular/forms';
import { cold } from 'jest-marbles';
import { createUser } from 'libs/admin/user-shared/test/user';
import { singleCold } from 'libs/tech-shared/test/marbles';
import { of, throwError } from 'rxjs';
import { UserAddFormservice } from './user-add.formservice';
import SpyInstance = jest.SpyInstance;

describe('UserAddFormService', () => {
  let formService: UserAddFormservice;
  let roleGroup: UntypedFormGroup;
  let alfaGroup: UntypedFormGroup;
  let administrationGroup: UntypedFormGroup;

  let userService: Mock<UserService>;
  let navigationService: Mock<NavigationService>;
  let snackBarService: Mock<SnackBarService>;

  beforeEach(() => {
    userService = { ...mock(UserService), refresh: jest.fn() };
    navigationService = mock(NavigationService);
    snackBarService = mock(SnackBarService);

    TestBed.configureTestingModule({
      providers: [
        UserAddFormservice,
        UntypedFormBuilder,
        { provide: UserService, useValue: userService },
        { provide: NavigationService, useValue: navigationService },
        { provide: SnackBarService, useValue: snackBarService },
      ],
    });

    formService = TestBed.inject(UserAddFormservice);
    roleGroup = <UntypedFormGroup>formService.form.get(UserAddFormservice.ROLLEN_GROUP);
    alfaGroup = <UntypedFormGroup>roleGroup.get(UserAddFormservice.ALFA_GROUP);
    administrationGroup = <UntypedFormGroup>roleGroup.get(UserAddFormservice.ADMINISTRATION_GROUP);
  });

  it('should create', () => {
    expect(formService).toBeTruthy();
  });

  describe('initAlfaGroupLogic', () => {
    it('should call handleAlfaGroupChange when value of form element changes', fakeAsync(() => {
      const handleAlfaGroupChangeSpy: SpyInstance = jest.spyOn(formService as any, 'handleAlfaGroupChange');

      alfaGroup.get(UserAddFormservice.LOESCHEN).setValue(true);

      tick();

      expect(handleAlfaGroupChangeSpy).toHaveBeenCalled();
    }));
  });

  describe('roleValidator', () => {
    it('should return error if no role is selected', () => {
      const result = formService.roleValidator()(roleGroup);

      expect(result).toEqual({ atLeastOneRoleSelected: true });
    });

    it('should return null if at least one role is selected', () => {
      alfaGroup.get(UserAddFormservice.LOESCHEN).setValue(true);

      const result = formService.roleValidator()(roleGroup);

      expect(result).toBeNull();
    });
  });

  describe('handleAlfaGroupChange', () => {
    it('should call disableUncheckedCheckboxes if any checkbox is checked', () => {
      jest.spyOn(formService as any, 'isAnyChecked').mockReturnValue(true);
      const disableUncheckedCheckboxesSpy: SpyInstance = jest.spyOn(formService as any, 'disableUncheckedCheckboxes');

      formService.handleAlfaGroupChange(alfaGroup);

      expect(disableUncheckedCheckboxesSpy).toHaveBeenCalled();
    });

    it('should call enableAllCheckboxes if not any checkbox is checked', () => {
      jest.spyOn(formService as any, 'isAnyChecked').mockReturnValue(false);
      const enableAllCheckboxesSpy: SpyInstance = jest.spyOn(formService as any, 'enableAllCheckboxes');

      formService.handleAlfaGroupChange(alfaGroup);

      expect(enableAllCheckboxesSpy).toHaveBeenCalled();
    });
  });

  describe('isAnyChecked', () => {
    it('should return false if no checkbox is checked', () => {
      const result = formService.isAnyChecked(alfaGroup);

      expect(result).toBe(false);
    });

    it('should return true if any checkbox is checked', () => {
      alfaGroup.get(UserAddFormservice.LOESCHEN).setValue(true);

      const result = formService.isAnyChecked(alfaGroup);

      expect(result).toBe(true);
    });
  });

  describe('disableUncheckedCheckboxes', () => {
    it('if control value is false then control should be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserAddFormservice.LOESCHEN);
      control.setValue(false);

      formService.disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(true);
    });

    it('if control value is true then control should NOT be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserAddFormservice.LOESCHEN);
      control.setValue(true);

      formService.disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(false);
    });
  });

  describe('updateCheckboxStates', () => {
    it('if control value is false then control should be disabled', () => {
      const control: AbstractControl = alfaGroup.get(UserAddFormservice.LOESCHEN);
      control.setValue(false);

      formService.disableUncheckedCheckboxes(alfaGroup);

      expect(control.disabled).toBe(true);
    });
  });

  describe('enableAllCheckboxes', () => {
    it('if control value is true then control should be enabled', () => {
      const control: AbstractControl = alfaGroup.get(UserAddFormservice.LOESCHEN);
      const enableSpy = jest.spyOn(control, 'enable');

      formService.enableAllCheckboxes(alfaGroup);

      expect(enableSpy).toHaveBeenCalled();
    });
  });

  describe('doSubmit', () => {
    const user: User = createUser();

    beforeEach(() => {
      userService.createInKeycloak.mockReturnValue(of(user));
    });

    it('should call createUser', () => {
      const createUserSpy: SpyInstance = jest.spyOn(formService as any, 'createUser');

      formService.submit();

      expect(createUserSpy).toHaveBeenCalled();
    });

    it('should call userService createInKeycloak', () => {
      formService.submit();

      expect(userService.createInKeycloak).toHaveBeenCalled();
    });

    it('should return user as stateResource', fakeAsync(() => {
      userService.createInKeycloak.mockReturnValue(singleCold(user, '-a'));

      const result = formService.submit();

      expect(result).toBeObservable(
        cold('ab', {
          a: createEmptyStateResource<User>(true),
          b: createStateResource(user),
        }),
      );
    }));

    it('should call handleOnCreateUserSuccess', fakeAsync(() => {
      const handleOnCreateUserSuccessSpy: SpyInstance = jest.spyOn(formService, 'handleOnCreateUserSuccess');

      formService.submit().subscribe();
      tick();

      expect(handleOnCreateUserSuccessSpy).toHaveBeenCalled();
    }));

    it('should call handleSubmitError on error', fakeAsync(() => {
      userService.createInKeycloak.mockReturnValue(throwError(() => new Error()));
      const handleSubmitErrorSpy: SpyInstance = jest.spyOn(formService, 'handleSubmitError');

      formService.submit().subscribe();
      tick();

      expect(handleSubmitErrorSpy).toHaveBeenCalled();
    }));
  });

  describe('handleOnCreateUserSuccess', () => {
    it('should show success message', () => {
      formService.handleOnCreateUserSuccess();

      expect(snackBarService.showInfo).toHaveBeenCalledWith('Der Benutzer wurde hinzugefügt.');
    });

    it('should navigate back to user list', () => {
      formService.handleOnCreateUserSuccess();

      expect(navigationService.navigate).toHaveBeenCalledWith(ROUTES.BENUTZER_UND_ROLLEN);
    });

    it('should refresh userList', () => {
      formService.handleOnCreateUserSuccess();

      expect(userService.refresh).toHaveBeenCalled();
    });
  });

  describe('handleSubmitError', () => {
    it('should return empty stateResource', () => {
      const result = formService.handleSubmitError();

      expect(result).toBeObservable(cold('(a|)', { a: createEmptyStateResource<User>() }));
    });

    it('should show error message', fakeAsync(() => {
      formService.handleSubmitError().subscribe();
      tick();

      expect(snackBarService.showError).toHaveBeenCalledWith('Der Benutzer konnte nicht hinzugefügt werden.');
    }));
  });

  describe('getRoles', () => {
    it('should return no roles when none are active', () => {
      const result: string[] = formService.getRoles(UserAddFormservice.ALFA_GROUP);

      expect(result).toEqual([]);
    });

    it('should return poststelle role when active', () => {
      alfaGroup.get(UserAddFormservice.POSTSTELLE).setValue(true);

      const result: string[] = formService.getRoles(UserAddFormservice.ALFA_GROUP);

      expect(result).toEqual([UserAddFormservice.POSTSTELLE]);
    });
  });
});
