/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { ROUTES } from '@admin-client/shared';
import { User, UserService } from '@admin-client/user-shared';
import { NavigationService } from '@alfa-client/navigation-shared';
import {
  AbstractFormService,
  createEmptyStateResource,
  createStateResource,
  EMPTY_STRING,
  StateResource,
} from '@alfa-client/tech-shared';
import { SnackBarService } from '@alfa-client/ui';
import { Injectable } from '@angular/core';
import {
  AbstractControl,
  FormControl,
  UntypedFormBuilder,
  UntypedFormGroup,
  ValidationErrors,
  ValidatorFn,
  Validators,
} from '@angular/forms';
import { catchError, map, Observable, of, startWith, tap } from 'rxjs';

@Injectable()
export class UserAddFormservice extends AbstractFormService<User> {
  public static readonly VORNAME: string = 'vorname';
  public static readonly NACHNAME: string = 'nachname';
  public static readonly BENUTZERNAME: string = 'benutzername';
  public static readonly EMAIL: string = 'e-mail';

  public static readonly ROLLEN_GROUP: string = 'rollen';
  public static readonly ADMINISTRATION_GROUP: string = 'administration';
  public static readonly ADMIN: string = 'ADMIN_ADMIN';
  public static readonly ALFA_GROUP: string = 'alfa';
  public static readonly LOESCHEN: string = 'VERWALTUNG_LOESCHEN';
  public static readonly USER: string = 'VERWALTUNG_USER';
  public static readonly POSTSTELLE: string = 'VERWALTUNG_POSTSTELLE';

  public static readonly USER_ADD_PREFIX: string = 'userAdd';

  constructor(
    public formBuilder: UntypedFormBuilder,
    private userService: UserService,
    private navigationService: NavigationService,
    private snackBarService: SnackBarService,
  ) {
    super(formBuilder);
    this.initAlfaGroupLogic();
  }

  protected initForm(): UntypedFormGroup {
    return this.formBuilder.group({
      [UserAddFormservice.VORNAME]: new FormControl(EMPTY_STRING, Validators.required),
      [UserAddFormservice.NACHNAME]: new FormControl(EMPTY_STRING, Validators.required),
      [UserAddFormservice.BENUTZERNAME]: new FormControl(EMPTY_STRING, Validators.required),
      [UserAddFormservice.EMAIL]: new FormControl(EMPTY_STRING, [Validators.required, Validators.email]),
      [UserAddFormservice.ROLLEN_GROUP]: this.formBuilder.group(
        {
          [UserAddFormservice.ADMINISTRATION_GROUP]: this.formBuilder.group({
            [UserAddFormservice.ADMIN]: new FormControl(false),
          }),
          [UserAddFormservice.ALFA_GROUP]: this.formBuilder.group({
            [UserAddFormservice.LOESCHEN]: new FormControl(false),
            [UserAddFormservice.USER]: new FormControl(false),
            [UserAddFormservice.POSTSTELLE]: new FormControl(false),
          }),
        },
        { validators: this.roleValidator() },
      ),
    });
  }

  roleValidator(): ValidatorFn {
    return (control: AbstractControl<UntypedFormGroup>): ValidationErrors | null => {
      const rolesGroups: UntypedFormGroup = control.value;

      if (this.anyRoleSelected(rolesGroups)) {
        return null;
      }

      return { atLeastOneRoleSelected: true };
    };
  }

  private anyRoleSelected(rolesGroups: UntypedFormGroup): boolean {
    for (const subGroup of Object.values(rolesGroups)) {
      if (Object.values(subGroup).includes(true)) {
        return true;
      }
    }
    return false;
  }

  protected initAlfaGroupLogic(): void {
    const alfaGroup: UntypedFormGroup = this.getRoleGroup(UserAddFormservice.ALFA_GROUP);
    alfaGroup.valueChanges.subscribe(() => {
      this.handleAlfaGroupChange(alfaGroup);
    });
  }

  handleAlfaGroupChange(group: UntypedFormGroup): void {
    const anyChecked: boolean = this.isAnyChecked(group);
    if (anyChecked) {
      this.disableUncheckedCheckboxes(group);
    } else {
      this.enableAllCheckboxes(group);
    }
  }

  isAnyChecked(group: UntypedFormGroup): boolean {
    return Object.keys(group.controls).some((key) => group.controls[key].value);
  }

  disableUncheckedCheckboxes(alfaGroup: UntypedFormGroup): void {
    for (const control of Object.values<AbstractControl>(alfaGroup.controls)) {
      if (control.value === false) control.disable({ emitEvent: false });
    }
  }

  enableAllCheckboxes(group: UntypedFormGroup): void {
    for (const control of Object.values<AbstractControl>(group.controls)) {
      control.enable({ emitEvent: false });
    }
  }

  protected doSubmit(): Observable<StateResource<User>> {
    const user: Partial<User> = this.createUser();
    return this.userService.createInKeycloak(user).pipe(
      map((createdUser: User): StateResource<User> => createStateResource(createdUser)),
      tap(() => this.handleOnCreateUserSuccess()),
      catchError((): Observable<StateResource<User>> => this.handleSubmitError()),
      startWith(createEmptyStateResource<User>(true)),
    );
  }

  handleOnCreateUserSuccess(): void {
    this.snackBarService.showInfo('Der Benutzer wurde hinzugefügt.');
    this.navigationService.navigate(ROUTES.BENUTZER_UND_ROLLEN);
    this.userService.refresh();
  }

  handleSubmitError(): Observable<StateResource<User>> {
    this.snackBarService.showError('Der Benutzer konnte nicht hinzugefügt werden.');
    return of(createEmptyStateResource<User>());
  }

  createUser(): Partial<User> {
    return {
      email: this.form.get(UserAddFormservice.EMAIL).value,
      username: this.form.get(UserAddFormservice.BENUTZERNAME).value,
      firstName: this.form.get(UserAddFormservice.VORNAME).value,
      lastName: this.form.get(UserAddFormservice.NACHNAME).value,
      enabled: true,
      clientRoles: {
        alfa: this.getRoles(UserAddFormservice.ALFA_GROUP),
        admin: this.getRoles(UserAddFormservice.ADMINISTRATION_GROUP),
      },
    };
  }

  getRoles(roleGroup: string): string[] {
    return this.getActiveRoles(this.getRoleGroup(roleGroup));
  }

  private getActiveRoles(formGroup: UntypedFormGroup): string[] {
    return Object.keys(formGroup.controls).filter((key) => formGroup.get(key).value);
  }

  private getRoleGroup(roleGroup: string): UntypedFormGroup {
    return <UntypedFormGroup>this.form.get(UserAddFormservice.ROLLEN_GROUP).get(roleGroup);
  }

  protected getPathPrefix(): string {
    return UserAddFormservice.USER_ADD_PREFIX;
  }
}
