/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { UserProfileResource } from '@alfa-client/user-profile-shared';
import { AuthConfig, OAuthEvent, OAuthService } from 'angular-oauth2-oidc';
import { Environment } from 'libs/environment-shared/src/lib/environment.model';
import { createUserProfileResource } from 'libs/user-profile-shared/test/user-profile';
import { Subject } from 'rxjs';
import { createEnvironment } from '../../../environment-shared/test/environment';
import { createAuthConfig, createOAuthEvent } from '../../test/authentication';
import { AuthenticationService } from './authentication.service';

describe('AuthenticationService', () => {
  let service: AuthenticationService;
  let oAuthService: Mock<OAuthService>;
  let environmentConfig: Environment;

  let eventsSubject: Subject<OAuthEvent>;

  beforeEach(() => {
    eventsSubject = new Subject<OAuthEvent>();

    oAuthService = <any>{
      ...mock(OAuthService),
      loadDiscoveryDocumentAndLogin: jest.fn().mockResolvedValue(() => Promise.resolve()),
      hasValidAccessToken: jest.fn(),
      hasValidIdToken: jest.fn(),
    };
    Object.defineProperty(oAuthService, 'events', { get: () => eventsSubject });

    environmentConfig = createEnvironment();
    service = new AuthenticationService(useFromMock(oAuthService), environmentConfig);
  });

  describe('login', () => {
    beforeEach(() => {
      service.buildAuthEventPromise = jest.fn();
    });

    it('should configure service with authConfig', async () => {
      const authConfig: AuthConfig = createAuthConfig();
      service.buildConfiguration = jest.fn().mockReturnValue(authConfig);

      await service.login();

      expect(oAuthService.configure).toHaveBeenCalledWith(authConfig);
    });

    it('should setup automatic silent refresh', async () => {
      await service.login();

      expect(oAuthService.setupAutomaticSilentRefresh).toHaveBeenCalled();
    });

    it('should set tokenValidator', async () => {
      oAuthService.tokenValidationHandler = null;

      await service.login();

      expect(oAuthService.tokenValidationHandler).not.toBeNull();
    });

    it('should build auth event promise', async () => {
      service.buildAuthEventPromise = jest.fn().mockResolvedValue(() => Promise.resolve());

      await service.login();

      expect(service.buildAuthEventPromise).toHaveBeenCalled();
    });

    it('should load discovery document and login', () => {
      service.login();

      expect(oAuthService.loadDiscoveryDocumentAndLogin).toHaveBeenCalled();
    });

    it('should return eventPromise', async () => {
      const promise: Promise<void> = Promise.resolve();
      service.buildAuthEventPromise = jest.fn().mockResolvedValue(promise);

      const returnPromise: Promise<void> = service.login();

      await expect(returnPromise).resolves.toBeUndefined();
    });
  });

  describe('build auth event promise', () => {
    const event: OAuthEvent = createOAuthEvent();

    beforeEach(() => {
      service.shouldProceedByAuthEvent = jest.fn().mockReturnValue(true);
      service.setCurrentUser = jest.fn();
      service.unsubscribeEvents = jest.fn();
    });

    it('should call shouldProceedByAuthEvent on event trigger', () => {
      service.buildAuthEventPromise();
      eventsSubject.next(event);

      expect(service.shouldProceedByAuthEvent).toHaveBeenCalledWith(event);
    });

    describe('on next', () => {
      it('should set current user', () => {
        service.buildAuthEventPromise();
        eventsSubject.next(event);

        expect(service.setCurrentUser).toHaveBeenCalled();
      });

      it('should unsubscribe event', () => {
        service.buildAuthEventPromise();
        eventsSubject.next(event);

        expect(service.unsubscribeEvents).toHaveBeenCalled();
      });

      it('should resolved promise with a valid event', async () => {
        const promise: Promise<void> = service.buildAuthEventPromise();
        eventsSubject.next(event);

        await expect(promise).resolves.toBeUndefined();
      });
    });

    describe('on error', () => {
      const errorMessage: string = 'Test Error';
      const error: Error = new Error(errorMessage);

      it('should unsubscribe event', () => {
        service.buildAuthEventPromise();
        eventsSubject.error(error);

        expect(service.unsubscribeEvents).toHaveBeenCalled();
      });

      it('should reject the promise with an error', async () => {
        const promise: Promise<void> = service.buildAuthEventPromise();

        eventsSubject.error(error);

        await expect(promise).rejects.toThrow(errorMessage);
      });
    });
  });

  describe('should proceed by auth event', () => {
    const event: OAuthEvent = createOAuthEvent();

    it('should call considered as login event', () => {
      service.consideredAsLoginEvent = jest.fn();

      service.shouldProceedByAuthEvent(event);

      expect(service.consideredAsLoginEvent).toHaveBeenCalledWith(event.type);
    });

    it('should return true on login event', () => {
      service.consideredAsLoginEvent = jest.fn().mockReturnValue(true);

      const proceed: boolean = service.shouldProceedByAuthEvent(event);

      expect(proceed).toBeTruthy();
    });

    it('should call considered as page reload event', () => {
      service.consideredAsLoginEvent = jest.fn().mockReturnValue(false);
      service.consideredAsPageReloadEvent = jest.fn();

      service.shouldProceedByAuthEvent(event);

      expect(service.consideredAsPageReloadEvent).toHaveBeenCalledWith(event.type);
    });

    it('should return true on page reload event', () => {
      service.consideredAsLoginEvent = jest.fn().mockReturnValue(false);
      service.consideredAsPageReloadEvent = jest.fn().mockReturnValue(true);

      const proceed: boolean = service.shouldProceedByAuthEvent(event);

      expect(proceed).toBeTruthy();
    });

    it('should return false on non login or page reload event', () => {
      service.consideredAsLoginEvent = jest.fn().mockReturnValue(false);
      service.consideredAsPageReloadEvent = jest.fn().mockReturnValue(false);

      const proceed: boolean = service.shouldProceedByAuthEvent(event);

      expect(proceed).toBeFalsy();
    });
  });

  describe('consideredAsLoginEvent', () => {
    it('should return true if event is "token_received"', () => {
      const event: string = 'token_received';

      const result: boolean = service.consideredAsLoginEvent(event);

      expect(result).toBeTruthy();
    });

    it('should return false if event is not "token_received"', () => {
      const event: string = 'something_else';

      const result: boolean = service.consideredAsLoginEvent(event);

      expect(result).toBeFalsy();
    });
  });

  describe('consideredAsPageReloadEvent', () => {
    it('should return true if event is "discovery_document_loaded" and tokens are valid', () => {
      service.hasValidToken = jest.fn().mockReturnValue(true);
      const event: string = 'discovery_document_loaded';

      const result: boolean = service.consideredAsPageReloadEvent(event);

      expect(result).toBeTruthy();
    });

    it('should return false if event is "discovery_document_loaded" and tokens are invalid', () => {
      service.hasValidToken = jest.fn().mockReturnValue(false);
      const event: string = 'discovery_document_loaded';

      const result: boolean = service.consideredAsPageReloadEvent(event);

      expect(result).toBeFalsy();
    });

    it('should return false if event is not "discovery_document_loaded" and tokens are valid', () => {
      service.hasValidToken = jest.fn().mockReturnValue(true);
      const event: string = 'something_else';

      const result: boolean = service.consideredAsPageReloadEvent(event);

      expect(result).toBeFalsy();
    });

    it('should return false if event is not "discovery_document_loaded" and tokens are invalid', () => {
      service.hasValidToken = jest.fn().mockReturnValue(false);
      const event: string = 'something_else';

      const result: boolean = service.consideredAsPageReloadEvent(event);

      expect(result).toBeFalsy();
    });
  });

  describe('hasValidToken', () => {
    it('should return true if both tokens are valid', () => {
      oAuthService.hasValidAccessToken.mockReturnValue(true);
      oAuthService.hasValidIdToken.mockReturnValue(true);

      const result: boolean = service.hasValidToken();

      expect(result).toBeTruthy();
    });

    it('should return false if access token is invalid', () => {
      oAuthService.hasValidAccessToken.mockReturnValue(false);

      const result: boolean = service.hasValidToken();

      expect(result).toBeFalsy();
    });

    it('should return false if id token is invalid', () => {
      oAuthService.hasValidAccessToken.mockReturnValue(true);
      oAuthService.hasValidIdToken.mockReturnValue(false);

      const result: boolean = service.hasValidToken();

      expect(result).toBeFalsy();
    });

    it('should return false if both tokens are invalid', () => {
      oAuthService.hasValidAccessToken.mockReturnValue(false);
      oAuthService.hasValidIdToken.mockReturnValue(false);

      const result: boolean = service.hasValidToken();

      expect(result).toBeFalsy();
    });
  });

  describe('set current user', () => {
    const identityClaims: Record<string, any> = {
      ['given_name']: 'Marco',
      ['family_name']: 'Polo',
    };

    beforeEach(() => {
      oAuthService.getIdentityClaims.mockReturnValue(identityClaims);
    });
    it('should call oAuthservice to get claims', () => {
      service.setCurrentUser();

      expect(oAuthService.getIdentityClaims).toHaveBeenCalled();
    });

    it('should update currentUser', () => {
      service.setCurrentUser();

      expect(service.currentUserResource.firstName).toEqual('Marco');
      expect(service.currentUserResource.lastName).toEqual('Polo');
    });
  });

  describe('getCurrentUserInitials', () => {
    it('should return currentUserResource', () => {
      const userProfile: UserProfileResource = {
        ...createUserProfileResource(),
        firstName: 'Marco',
        lastName: 'Polo',
      };
      service.currentUserResource = userProfile;

      const currentUserInitials: string = service.getCurrentUserInitials();

      expect(currentUserInitials).toEqual('MP');
    });
  });

  describe('logout', () => {
    it('should call oAuthService revokeTokenAndLogout', () => {
      service.logout();

      expect(oAuthService.revokeTokenAndLogout).toHaveBeenCalled();
    });
  });
});
