/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import { InteractivityChecker } from '@angular/cdk/a11y';
import { ComponentRef, ElementRef, Renderer2, ViewContainerRef } from '@angular/core';
import { TestBed } from '@angular/core/testing';
import { faker } from '@faker-js/faker/.';
import { TooltipComponent } from './tooltip.component';
import { TooltipDirective, TooltipPosition } from './tooltip.directive';

class MockElementRef extends ElementRef {
  nativeElement = {
    contains: jest.fn(),
    appendChild: jest.fn(),
  };
}

describe('TooltipDirective', () => {
  let directive: TooltipDirective;
  const mockComponentRef: ComponentRef<TooltipComponent> = {
    setInput: jest.fn(),
    destroy: jest.fn(),
    onDestroy: jest.fn(),
    componentType: TooltipComponent,
    changeDetectorRef: null,
    location: null,
    hostView: null,
    injector: null,
    instance: { id: '', left: 0, top: 0, text: '', show: false, position: TooltipPosition.BELOW, class: '', leftOffset: 0 },
  };
  const parentRect: DOMRect = { bottom: 0, top: 0, height: 0, left: 0, right: 0, toJSON: jest.fn(), width: 0, x: 0, y: 0 };

  beforeEach((): void => {
    TestBed.configureTestingModule({
      providers: [ViewContainerRef, { provide: ElementRef, useClass: MockElementRef }, Renderer2, InteractivityChecker],
    });
    TestBed.runInInjectionContext(() => {
      directive = new TooltipDirective();
    });
  });

  it('should create a directive', () => {
    expect(directive).toBeTruthy();
  });

  describe('set tooltip', () => {
    beforeEach(() => {
      directive.createTooltip = jest.fn();
      directive.destroy = jest.fn();
    });

    it('should destroy tooltip if it exists already', () => {
      directive.componentRef = mockComponentRef;

      directive.tooltip = 'test';

      expect(directive.destroy).toHaveBeenCalled();
    });

    it('should not create tooltip if text is empty', () => {
      directive.tooltip = '';

      expect(directive.createTooltip).not.toHaveBeenCalled();
    });

    it('should create tooltip', () => {
      directive.tooltip = 'test';

      expect(directive.createTooltip).toHaveBeenCalled();
    });
  });

  describe('ngOnDestroy', () => {
    it('should destroy tooltip', () => {
      directive.destroy = jest.fn();

      directive.ngOnDestroy();

      expect(directive.destroy).toHaveBeenCalled();
    });
  });

  describe('createTooltip', () => {
    const tooltipText: string = faker.lorem.sentence();

    beforeEach(() => {
      directive.viewContainerRef.createComponent = jest.fn().mockReturnValue({ location: { nativeElement: {} } });
      directive.setAriaAttribute = jest.fn();
      directive.setInitialTooltipProperties = jest.fn();
      directive.getParentElement = jest.fn().mockReturnValue({ appendChild: jest.fn() });
      directive.interactivityChecker.isFocusable = jest.fn();
    });

    it('should create tooltip component', () => {
      directive.createTooltip(tooltipText);

      expect(directive.viewContainerRef.createComponent).toHaveBeenCalled();
    });

    it('should get parent element', () => {
      directive.createTooltip(tooltipText);

      expect(directive.getParentElement).toHaveBeenCalled();
    });

    it('should insert tooltip component to parent element', () => {
      directive.createTooltip(tooltipText);

      expect(directive.parentElement.appendChild).toHaveBeenCalled();
    });

    it('should set initial tooltip properties', () => {
      directive.createTooltip(tooltipText);

      expect(directive.setInitialTooltipProperties).toHaveBeenCalled();
    });

    it('should set aria attribute to parent', () => {
      directive.createTooltip(tooltipText);

      expect(directive.setAriaAttribute).toHaveBeenCalled();
    });

    it('should not create tooltip', () => {
      directive.createTooltip('');

      expect(directive.getParentElement).not.toHaveBeenCalled();
    });
  });

  describe('showTooltip', () => {
    beforeEach(() => {
      directive.componentRef = mockComponentRef;
      directive.setTooltipProperties = jest.fn();
      directive.elementRef.nativeElement.contains = jest.fn().mockReturnValue(true);
    });

    it('should not set tooltip properties if component ref is null', () => {
      directive.componentRef = null;

      directive.showTooltip();

      expect(directive.setTooltipProperties).not.toHaveBeenCalled();
    });

    it('should check if element focused', () => {
      directive.showTooltip();

      expect(directive.elementRef.nativeElement.contains).toHaveBeenCalled();
    });

    it('should set attachedToFocused field', () => {
      directive.showTooltip();

      expect(directive.attachedToFocused).toBeTruthy();
    });

    it('should set tooltip properties', () => {
      directive.showTooltip();

      expect(directive.setTooltipProperties).toHaveBeenCalled();
    });
  });

  describe('hideTooltip', () => {
    it('should hide tooltip', () => {
      directive.hide = jest.fn();

      directive.hideTooltip();

      expect(directive.hide).toHaveBeenCalled();
    });
  });

  describe('onKeydown', () => {
    it('should hide tooltip if escape key pressed', () => {
      directive.hide = jest.fn();
      const escapeEvent: KeyboardEvent = { ...new KeyboardEvent('esc'), key: 'Escape' };

      directive.onKeydown(escapeEvent);

      expect(directive.hide).toHaveBeenCalled();
    });
  });

  describe('setInitialTooltipProperties', () => {
    beforeEach(() => {
      directive.componentRef = mockComponentRef;
    });

    it('should set tooltip text', () => {
      directive.setInitialTooltipProperties('Test', 'tooltip-1');

      expect(directive.componentRef.instance.text).toBe('Test');
    });

    it('should set tooltip id', () => {
      directive.setInitialTooltipProperties('Test', 'tooltip-1');

      expect(directive.componentRef.instance.id).toBe('tooltip-1');
    });
  });

  describe('setTooltipProperties', () => {
    beforeEach(() => {
      directive.componentRef = mockComponentRef;
      directive.setTooltipOffsetAndPosition = jest.fn();
      directive.elementRef.nativeElement.getBoundingClientRect = jest.fn().mockReturnValue({ left: 0, right: 1000 });
      directive.getTopPosition = jest.fn().mockReturnValue(1000);
    });

    it('should get bounding client rect', () => {
      directive.setTooltipProperties();

      expect(directive.elementRef.nativeElement.getBoundingClientRect).toHaveBeenCalled();
    });

    it('should get top position for tooltip', () => {
      directive.setTooltipProperties();

      expect(directive.getTopPosition).toHaveBeenCalled();
    });

    it('should set tooltip instance properties', () => {
      directive.setTooltipProperties();

      expect(directive.componentRef.instance).toMatchObject({
        class: '',
        left: 500,
        top: 1000,
        show: true,
      });
    });
  });

  describe('setAutoPosition', () => {
    it('should set tooltip position above if it cannot be displayed below', () => {
      directive.tooltipPosition = TooltipPosition.BELOW;
      const rectForAbove: DOMRect = { ...parentRect, bottom: 980 };

      directive.setAutoPosition(36, rectForAbove, 1000);

      expect(directive.position).toBe(TooltipPosition.ABOVE);
    });

    it('should set tooltip position below if it cannot be displayed above', () => {
      directive.tooltipPosition = TooltipPosition.ABOVE;
      const rectForBelow: DOMRect = { ...parentRect, top: 15 };

      directive.setAutoPosition(36, rectForBelow, 1000);

      expect(directive.position).toBe(TooltipPosition.BELOW);
    });

    it('should set tooltip on selected position otherwise', () => {
      directive.tooltipPosition = TooltipPosition.ABOVE;
      const rectForStandard: DOMRect = { ...parentRect, top: 500 };

      directive.setAutoPosition(36, rectForStandard, 1000);

      expect(directive.position).toBe(TooltipPosition.ABOVE);
    });
  });

  describe('setLeftOffset', () => {
    beforeEach(() => {
      directive.leftOffset = 0;
    });

    it('should set 0 offset, if parent wider than tooltip', () => {
      const rectForStandard: DOMRect = { ...parentRect, left: 50, right: 500, width: 100 };

      directive.setLeftOffset(36, rectForStandard, 1000);

      expect(directive.leftOffset).toBe(0);
    });

    it('should set positive left offset', () => {
      const rectForLeft: DOMRect = { ...parentRect, left: 10 };

      directive.setLeftOffset(50, rectForLeft, 1000);

      expect(directive.leftOffset).toBe(15);
    });

    it('should set negative left offset', () => {
      const rectForRight: DOMRect = { ...parentRect, left: 50, right: 990 };

      directive.setLeftOffset(50, rectForRight, 1000);

      expect(directive.leftOffset).toBe(-15);
    });

    it('should set 0 offset', () => {
      const rectForStandard: DOMRect = { ...parentRect, left: 50, right: 500 };

      directive.setLeftOffset(36, rectForStandard, 1000);

      expect(directive.leftOffset).toBe(0);
    });
  });

  describe('setTooltipOffsetAndPosition', () => {
    beforeEach(() => {
      directive.setAutoPosition = jest.fn();
      directive.setLeftOffset = jest.fn();
      directive.componentRef = Object.assign(mockComponentRef, {
        location: {
          nativeElement: { children: [{ getBoundingClientRect: jest.fn().mockReturnValue({ height: 100, width: 200 }) }] },
        },
      });
      directive.elementRef.nativeElement.getBoundingClientRect = jest.fn().mockReturnValue({});
      window.innerHeight = 404;
      window.innerWidth = 503;
    });

    it('should set tooltip auto position', () => {
      directive.setTooltipOffsetAndPosition();

      expect(directive.setAutoPosition).toHaveBeenCalledWith(100, {}, 404);
    });

    it('should set tooltip left offset', () => {
      directive.setTooltipOffsetAndPosition();

      expect(directive.setLeftOffset).toHaveBeenCalledWith(200, {}, 503);
    });
  });

  describe('setAriaAttribute', () => {
    beforeEach(() => {
      directive.renderer.setAttribute = jest.fn();
      directive.parentElement = null;
      directive.tooltipId = 'test';
    });

    it('should set aria-describedby attribute', () => {
      directive.setAriaAttribute('aria-describedby');

      expect(directive.renderer.setAttribute).toHaveBeenCalledWith(null, 'aria-describedby', 'test');
    });

    it('should set aria-labelledby attribute', () => {
      directive.setAriaAttribute('aria-labelledby');

      expect(directive.renderer.setAttribute).toHaveBeenCalledWith(null, 'aria-labelledby', 'test');
    });
  });

  describe('removeAriaLabelledBy', () => {
    beforeEach(() => {
      directive.renderer.removeAttribute = jest.fn();
      directive.parentElement = null;
    });

    it('should remove aria-describedby attribute', () => {
      directive.removeAriaAttribute('aria-describedby');

      expect(directive.renderer.removeAttribute).toHaveBeenCalledWith(null, 'aria-describedby');
    });

    it('should remove aria-labelledby attribute', () => {
      directive.removeAriaAttribute('aria-labelledby');

      expect(directive.renderer.removeAttribute).toHaveBeenCalledWith(null, 'aria-labelledby');
    });
  });

  describe('getParentElement', () => {
    beforeEach(() => {
      directive.getFocusableElement = jest.fn();
      directive.interactivityChecker.isFocusable = jest.fn();
    });

    it('should check if element focusable', () => {
      const element = document.createElement('button');

      directive.getParentElement(element);

      expect(directive.interactivityChecker.isFocusable).toHaveBeenCalled();
    });

    it('should return element itself if it is focusable', () => {
      const element = document.createElement('button');
      directive.interactivityChecker.isFocusable = jest.fn().mockReturnValue(true);

      const result: HTMLElement = directive.getParentElement(element);

      expect(result).toBe(element);
    });

    it('should get focusable element from the given element', () => {
      const element = document.createElement('button');

      directive.getParentElement(element);

      expect(directive.getFocusableElement).toHaveBeenCalled();
    });

    it('should return focusable element if it is not null', () => {
      const element = document.createElement('div');
      const subElement = document.createElement('button');
      directive.getFocusableElement = jest.fn().mockReturnValue(subElement);

      const result: HTMLElement = directive.getParentElement(element);

      expect(result).toBe(subElement);
    });

    it('should return element itself if focusable element is null', () => {
      const element = document.createElement('div');
      directive.getFocusableElement = jest.fn().mockReturnValue(null);

      const result: HTMLElement = directive.getParentElement(element);

      expect(result).toBe(element);
    });
  });

  describe('getFocusableElement', () => {
    it('should return null', () => {
      const simpleElement = document.createElement('a');

      const result: HTMLElement = directive.getFocusableElement(simpleElement);

      expect(result).toBeNull();
    });

    it('should return focusable child element', () => {
      const nestedElement = document.createElement('div');
      const buttonElement = document.createElement('button');
      nestedElement.appendChild(buttonElement);

      const result: HTMLElement = directive.getFocusableElement(nestedElement);

      expect(result).toBe(buttonElement);
    });
  });

  describe('getTopPosition', () => {
    it('should return top for position above', () => {
      const result: number = directive.getTopPosition(123, 456, TooltipPosition.ABOVE, false);

      expect(result).toBe(123);
    });

    it('should return top for position above with offset', () => {
      const result: number = directive.getTopPosition(123, 456, TooltipPosition.ABOVE, true);

      expect(result).toBe(119);
    });

    it('should return top for position below', () => {
      const result: number = directive.getTopPosition(123, 456, TooltipPosition.BELOW, false);

      expect(result).toBe(456);
    });

    it('should return top for position below with offset', () => {
      const result: number = directive.getTopPosition(123, 456, TooltipPosition.BELOW, true);

      expect(result).toBe(460);
    });
  });

  describe('hide', () => {
    beforeEach(() => {
      directive.componentRef = Object.assign(mockComponentRef, { instance: { ...mockComponentRef.instance, show: true } });
    });

    it('should hide component', () => {
      directive.hide();

      expect(directive.componentRef.instance.show).toBeFalsy();
    });
  });

  describe('destroy', () => {
    beforeEach(() => {
      directive.componentRef = mockComponentRef;
      directive.removeAriaAttribute = jest.fn();
    });

    it('should set component ref to null', () => {
      directive.destroy();

      expect(directive.componentRef).toBeNull();
    });

    it('should remove aria attribute', () => {
      directive.destroy();

      expect(directive.removeAriaAttribute).toHaveBeenCalled();
    });

    it('should set focusable element to null', () => {
      directive.destroy();

      expect(directive.parentElement).toBeNull();
    });
  });
});
